# coding: utf-8
ERR = -1
OK = 0

ZSKIPLIST_MAXLEVEL = 32
ZSKIPLIST_P = 0.25

ZADD_NONE = 0
ZADD_INCR = 1 << 0
ZADD_NX = 1 << 1

ZADD_NOP = 1 << 3
ZADD_NAN = 1 << 4
ZADD_ADDED = 1 << 5
ZADD_UPDATED = 1 << 6



# coding: utf-8

from boost_collections.zskiplist.zskiplist_level import ZskiplistLevel


class ZskiplistNode(object):

    def __init__(self, level, score, ele=None):
        super(ZskiplistNode, self).__init__()
        self.ele = ele
        self.score = score
        self.backward = None
        self.level = [ZskiplistLevel() for _ in range(0, level)]



# coding: utf-8
from numpy.core.umath import isnan

from boost_collections.zskiplist.constant import ZSKIPLIST_MAXLEVEL
from boost_collections.zskiplist.util import elecmp, zsl_random_level
from boost_collections.zskiplist.zskiplist_node import ZskiplistNode


class Zskiplist(object):

    def __init__(self):
        super(Zskiplist, self).__init__()
        self.header = ZskiplistNode(ZSKIPLIST_MAXLEVEL, 0)
        for j in range(0, ZSKIPLIST_MAXLEVEL):
            self.header.level[j].forward = None
            self.header.level[j].span = 0
        self.header.backward = None
        self.tail = None
        self.length = 0
        self.level = 1

    def zsl_insert(self, score, ele):
        """
        Insert a new node in the skiplist. Assumes the element does not already
        exist (up to the caller to enforce that). The skiplist takes ownership
        of the passed string 'ele'.
        :param score:
        :param ele:
        :return:
        """
        update = [None] * ZSKIPLIST_MAXLEVEL
        rank = [0] * ZSKIPLIST_MAXLEVEL

        assert (not isnan(score)), 'score can not be NaN'
        x = self.header

        i = self.level - 1
        while i >= 0:
            rank[i] = 0 if i == (self.level - 1) else rank[i + 1]
            while (x.level[i].forward and
                   (x.level[i].forward.score < score or
                    (x.level[i].forward.score == score and
                     elecmp(x.level[i].forward.ele, ele) < 0))):
                rank[i] += x.level[i].span
                x = x.level[i].forward
            update[i] = x
            i -= 1

        level = zsl_random_level()
        if level > self.level:
            for i in range(self.level, level):
                rank[i] = 0
                update[i] = self.header
                update[i].level[i].span = self.length
            self.level = level
        x = ZskiplistNode(level=level, score=score, ele=ele)
        for i in range(0, level):
            x.level[i].forward = update[i].level[i].forward
            update[i].level[i].forward = x
            x.level[i].span = update[i].level[i].span - (rank[0] - rank[i])
            update[i].level[i].span = (rank[0] - rank[i]) + 1

        for i in range(level, self.level):
            update[i].level[i].span += 1

        x.backward = None if (update[0] is self.header) else update[0]

        if x.level[0].forward:
            x.level[0].forward.backward = x
        else:
            self.tail = x
        self.length += 1
        return x

    def zsl_delete_node(self, x, update):
        for i in range(0, self.level):
            if update[i].level[i].forward is x:
                update[i].level[i].span += x.level[i].span - 1
                update[i].level[i].forward = x.level[i].forward
            else:
                update[i].level[i].span -= 1
        if x.level[0].forward:
            x.level[0].forward.backward = x.backward
        else:
            self.tail = x.backward
        while self.level > 1 and self.header.level[self.level - 1].forward is None:
            self.level -= 1
        self.length -= 1

    def zsl_delete(self, score, ele):
        update = [None] * ZSKIPLIST_MAXLEVEL
        x = self.header
        i = self.level - 1
        while i >= 0:
            while (x.level[i].forward and
                   (x.level[i].forward.score < score or
                    (x.level[i].forward.score == score and
                     elecmp(x.level[i].forward.ele, ele) < 0))):
                x = x.level[i].forward
            update[i] = x
            i -= 1
        x = x.level[0].forward
        if x and score == x.score and elecmp(x.ele, ele) == 0:
            self.zsl_delete_node(x=x, update=update)
            return 1, x
        return 0, None

    def zsl_get_element_by_rank(self, rank):
        """
        Finds an element by its rank. The rank argument needs to be 1-based.
        :param rank:
        :return:
        """
        traversed = 0
        x = self.header
        i = self.level - 1
        while i >= 0:
            while x.level[i].forward and (traversed + x.level[i].span) <= rank:
                traversed += x.level[i].span
                x = x.level[i].forward
            if traversed == rank:
                return x
            i -= 1
        return None



# coding: utf-8
from numpy import random

from boost_collections.zskiplist.constant import ZSKIPLIST_P, ZSKIPLIST_MAXLEVEL


def zsl_random_level():
    level = 1
    while random.rand() < ZSKIPLIST_P:
        level += 1

    return level if (level < ZSKIPLIST_MAXLEVEL) else ZSKIPLIST_MAXLEVEL


def elecmp(s1, s2):
    if s1 == s2:
        return 0
    elif s1 > s2:
        return 1
    else:
        return -1



# coding: utf-8
from numpy.core.umath import isnan

from boost_collections.zskiplist.constant import ZADD_INCR, ZADD_NX, ZADD_NAN, ZADD_NOP, ZADD_UPDATED, ZADD_ADDED, OK, \
    ERR
# from boost_collections.zskiplist.zskiplist import Zskiplist
from boost_collections.zskiplist.zskiplist_ex import ZskiplistEx


class Zset(object):

    def __init__(self):
        super(Zset, self).__init__()
        self.dict_ = dict()
        # self.zsl = Zskiplist()
        self.zsl = ZskiplistEx()

    def zset_add(self, score, ele, flags):
        """
        Add a new element or update the score of an existing element in a sorted
        set.
 
        The set of flags change the command behavior. They are passed with an integer
        and will return other flags to indicate different conditions.

        The input flags are the following:

        ZADD_INCR: Increment the current element score by 'score' instead of updating
                 the current element score. If the element does not exist, we
                 assume 0 as previous score.
        ZADD_NX:   Perform the operation only if the element does not exist.

        When ZADD_INCR is used, the new score of the element is returned with result.

        The returned flags are the following:

        ZADD_NAN:     The resulting score is not a number.
        ZADD_ADDED:   The element was added (not present before the call).
        ZADD_UPDATED: The element score was updated.
        ZADD_NOP:     No operation was performed because of NX.

        Return value:

        The function returns 1 on success, and sets the appropriate flags
        ADDED or UPDATED to signal what happened during the operation (note that
        none could be set if we re-added an element using the same score it used
        to have, or in the case a zero increment is used).

        The function returns 0 on erorr, currently only when the increment
        produces a NAN condition, or when the 'score' value is NAN since the
        start.
        :param score:
        :param ele:
        :param flags:
        :return:
        """
        incr = (flags & ZADD_INCR) != 0
        nx = (flags & ZADD_NX) != 0

        if isnan(score):
            return 0, ZADD_NAN, 0
        dict_ = self.dict_
        de = dict_.get(ele, None)
        if de is not None:
            if nx:
                return 1, ZADD_NOP, 0
            curscore = de
            if incr:
                score += curscore
                if isnan(score):
                    return 0, ZADD_NAN, 0
            # Remove and re-insert when score changes.
            if score != curscore:
                result = self.zsl.zsl_delete(score=curscore, ele=ele)
                assert result == 1
                result = self.zsl.zsl_insert(score=score, ele=ele)
                assert result == 1
                dict_[ele] = score

            return 1, ZADD_UPDATED, score
        else:
            result = self.zsl.zsl_insert(score=score, ele=ele)
            assert result == 1
            assert dict_.setdefault(ele, score) == score
            return 1, ZADD_ADDED, score

    def zset_del(self, ele):
        """
        Delete the element 'ele' from the sorted set, returning 1 if the element
        existed and was deleted, 0 otherwise (the element was not there).
        :param ele:
        :return:
        """
        de = self.dict_.pop(ele, None)
        if de is not None:
            result = self.zsl.zsl_delete(score=de, ele=ele)
            assert result == 1
            return 1
        return 0

    def zset_length(self):
        """
        Return the length of the sorted set
        :return:
        """
        length = self.zsl.zsl_length()
        return length

    def zset_score(self, member):
        """
        Return (by reference) the score of the specified member of the sorted set.
        If the element does not exist ERR is returned
        otherwise OK and score is returned.
        If 'member' is NULL, ERR is returned.
        :param member:
        :return:
        """
        if member is None:
            return ERR, 0
        de = self.dict_.get(member, None)
        if de is None:
            return ERR, 0
        return OK, de

    def zset_get_floor_element_by_score(self, score):
        """
        Returns a key-value mapping associated with the greatest key
        less than or equal to the given key,or None if there is no such key.
        :param score:
        :return:
        """
        return self.zsl.zsl_get_floor_element_by_score(score)

    def zset_get_lower_element_by_score(self, score):
        """
        Returns a key-value mapping associated with the greatest key
        less than the given key,or None if there is no such key.
        :param score:
        :return:
        """
        return self.zsl.zsl_get_lower_element_by_score(score)



import os

from setuptools import setup, find_packages, Extension


def fpath(name):
    return os.path.join(os.path.dirname(__file__), name)


def read(fname):
    return open(fpath(fname)).read()


def desc():
    try:
        return read('README.md')
    except:
        return ''


from setuptools.command.build_ext import build_ext as _build_ext


class BuildExtWithoutPlatformSuffix(_build_ext):

    def get_ext_filename(self, ext_name):
        ext_path = ext_name.split('.')
        ext_suffix = '.so'
        return os.path.join(*ext_path) + ext_suffix


setup(
    name='ratel',
    keywords='ratel',
    version='0.5.0',
    description='Sorted data structure implementation by using skiplist and dict with Python and C',
    long_description=desc(),
    long_description_content_type="text/markdown",
    url='https://github.com/524243642/ratel',
    license='MIT',
    author='Bin Zhang',
    author_email='524243642@qq.com',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    packages=find_packages(exclude=['tests*']),
    test_suite='tests',
    install_requires=[
        'numpy>=1.11.3'
    ],
    ext_modules=[
        Extension(
            'boost_collections.zskiplist._zskiplist',
            extra_compile_args=['-Wall'],
            sources=['boost_collections/zskiplist/t_zset.c']
        ),
    ],
    cmdclass={'build_ext': BuildExtWithoutPlatformSuffix},
)
